//----------------------------------------------------------------------------
// mas.c
//
// Coded by Jagos (a.k.a. Jacobo Romero Manrique)
// jacoboromero@terra.es
//
// Greetings to Debugger and HexDump.
//----------------------------------------------------------------------------
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

typedef signed char s8;
typedef signed short s16;
typedef signed int s32;

typedef int fixed;

#define itofix(a)	(a<<16)
#define fixtoi(a)	(a>>16)

#define DISPCNT		((u16 *)0x04000000)
#define VCOUNT		((volatile u8 *)0x04000006);
#define PALETTE		((u16 *)0x05000000)
#define FRAME0		((u8 *)0x06000000)
#define FRAME1		((u8 *)0x0600A000)

#define SCRWIDTH	240
#define SCRHEIGHT	160
#define BMPWIDTH	64
#define BMPHEIGHT	64
#define	MAXANGLE	1023

extern u8 bitmap[BMPHEIGHT][BMPWIDTH];
extern u16 palette[256];
extern s16 sintable[1024];
u8 *screen;

// In order to avoid divisions I have used the Bresenham's line algorithm a lot for
// the linear interpolations.
void putsquash(int x, int y1, int y2)
{
	int dx, dy, du, dv, da;
	int A, Ainc1, Ainc2, B, Binc1, Binc2, C, Cinc1, Cinc2;
	int radius, angle;
	int i, j, k;
	u8 *source, *dest;

	dy=y2-y1;
	du=BMPWIDTH;
	dv=BMPHEIGHT-1;
	da=MAXANGLE/2;		// Change this value for cool effects (try MAXANGLE*4 or so)

	angle=0;
	radius=(BMPWIDTH-(dy*BMPWIDTH/BMPHEIGHT))/2;

	if(dy>dv) {
		A=2*dv-dy;
		Ainc1=2*dv;
		Ainc2=2*(dv-dy);
	}
	else {
		A=2*dy-dv;
		Ainc1=2*dy;
		Ainc2=2*(dy-dv);
	}

	if(dy>da) {
		B=2*da-dy;
		Binc1=2*da;
		Binc2=2*(da-dy);
	}
	else {
		B=2*dy-da;
		Binc1=2*dy;
		Binc2=2*(dy-da);
	}

	source=(u8 *)bitmap;

	// Clip up
	if(y1<0) {
		for(i=-y1; i; i--) {
			if(dy>dv) {
				if(A<0)
					A+=Ainc1;
				else {
					A+=Ainc2;
					source+=BMPWIDTH;
				}
			}
			else {
				while(A<0) { source+=BMPWIDTH; A+=Ainc1; }
				A+=Ainc2;
				source+=BMPWIDTH;
			}

			if(dy>da) {
				if(B<0)
					B+=Binc1;
				else {
					B+=Binc2;
					angle++;
				}
			}
			else {
				while(B<0) { angle++; B+=Binc1; }
				B+=Binc2;
				angle++;
			}
		}
		y1=0;
	}

	dest=&screen[y1*SCRWIDTH + x];

	// Clip down
	if(y2>=SCRHEIGHT) y2=SCRHEIGHT-1;

	for(i=y1; i<=y2; i++) {

		k=BMPWIDTH/2+((radius*sintable[angle&MAXANGLE])>>15);
		dx=2*k;
		dest-=k;

		if(dx==du)
			for(j=dx; j; j--) *dest++=*source++;
		else if(dx>du) {
			C=2*du-dx;
			Cinc1=2*du;
			Cinc2=2*(du-dx);

			for(j=dx; j; j--) {
				*dest++=*source;
				if(C<0)
					C+=Cinc1;
				else {
					C+=Cinc2;
					source++;
				}
			}
		}
		else {
			C=2*dx-du;
			Cinc1=2*dx;
			Cinc2=2*(dx-du);

			for(j=dx; j; j--) {
				while(C<0) { C+=Cinc1; source++; }
				C+=Cinc2;
				*dest++=*source++;
			}
		}

		dest+=SCRWIDTH-k;

		if(dy>dv) {
			if(A<0) {
				A+=Ainc1;
				source-=BMPWIDTH;
			}
			else
				A+=Ainc2;
		}
		else {
			while(A<0) { source+=BMPWIDTH; A+=Ainc1; }
			A+=Ainc2;
		}

		if(dy>da) {
			if(B<0)
				B+=Binc1;
			else {
				B+=Binc2;
				angle++;
			}
		}
		else {
			while(B<0) { angle++; B+=Binc1; }
			B+=Binc2;
			angle++;
		}
	}
}

void flip()
{
	u8 line=*VCOUNT;

	while(line<160) line=*VCOUNT;

	if(screen==FRAME0)
		{ screen=FRAME1; *DISPCNT=0x0404; }
	else
		{ screen=FRAME0; *DISPCNT=0x0414; }
}

void clear()
{
	int i;
	u32 *p=(u32 *)screen;

	for(i=9600; i; i--) *p++=0;
}

void C_entry()
{
	fixed y, vy, ay, x, A;
	int angle, i;
	const int w=96;
	enum { PRESSING, BOINGBOING } state;

	*DISPCNT=0x0404;
	screen=FRAME1;
	for(i=0; i<256; i++) PALETTE[i]=palette[i];

	A=0;
	y=itofix(-BMPHEIGHT);
	vy=0;
	ay=itofix(1);

	state=PRESSING;

	for(;;) {
		// x = A*sin(wt)
		if(A>=6144) A-=6144;
		angle+=w;
		x=(fixtoi(A)*sintable[angle&MAXANGLE])>>15;

		vy+=ay; y+=vy;

		if(fixtoi(y)>=SCRHEIGHT-BMPHEIGHT+x-1) {
			y=itofix(SCRHEIGHT-BMPHEIGHT+x-1);

			if(state==PRESSING) {
				vy-=ay;
				if(!A) {
					A=itofix(BMPHEIGHT/4);
					angle=0;
				}
				else if(angle>=384) {
					state=BOINGBOING;
				}
			}
			else vy=-vy+(vy>>3);
		}

		clear();
		putsquash(120, fixtoi(y)+x, fixtoi(y)+BMPHEIGHT-x);
		flip();
	}
}